@@ -1,6 +1,6 @@
 /*
   SDL_image:  An example image loading library for use with SDL
-  Copyright (C) 1997-2016 Sam Lantinga <slouken@libsdl.org>
+  Copyright (C) 1997-2018 Sam Lantinga <slouken@libsdl.org>
 
   This software is provided 'as-is', without any express or implied
   warranty.  In no event will the authors be held liable for any damages
@@ -21,11 +21,6 @@
 
 /* This is a XCF image file loading framework */
 
-#include <stdio.h>
-#include <ctype.h>
-#include <string.h>
-#include <stdlib.h>
-
 #include "SDL_endian.h"
 #include "SDL_image.h"
 
@@ -251,6 +246,7 @@ static Uint32 Swap32 (Uint32 v) {
 }
 
 static void xcf_read_property (SDL_RWops * src, xcf_prop * prop) {
+  Uint32 len;
   prop->id = SDL_ReadBE32 (src);
   prop->length = SDL_ReadBE32 (src);
 
@@ -274,7 +270,12 @@ static void xcf_read_property (SDL_RWops * src, xcf_prop * prop) {
     break;
   case PROP_COMPRESSION:
   case PROP_COLOR:
-    SDL_RWread (src, &prop->data, prop->length, 1);
+    if (prop->length > sizeof(prop->data)) {
+        len = sizeof(prop->data);
+    } else {
+        len = prop->length;
+    }
+    SDL_RWread(src, &prop->data, len, 1);
     break;
   case PROP_VISIBLE:
     prop->data.visible = SDL_ReadBE32 (src);
@@ -298,6 +299,9 @@ static xcf_header * read_xcf_header (SDL_RWops * src) {
   xcf_prop prop;
 
   h = (xcf_header *) SDL_malloc (sizeof (xcf_header));
+  if (!h) {
+    return NULL;
+  }
   SDL_RWread (src, h->sign, 14, 1);
   h->width       = SDL_ReadBE32 (src);
   h->height      = SDL_ReadBE32 (src);
@@ -316,11 +320,22 @@ static xcf_header * read_xcf_header (SDL_RWops * src) {
       h->compr = (xcf_compr_type)prop.data.compression;
     else if (prop.id == PROP_COLORMAP) {
       // unused var: int i;
+      Uint32 cm_num;
+      unsigned char *cm_map;
 
-      h->cm_num = prop.data.colormap.num;
-      h->cm_map = (unsigned char *) SDL_malloc (sizeof (unsigned char) * 3 * h->cm_num);
-      SDL_memcpy (h->cm_map, prop.data.colormap.cmap, 3*sizeof (char)*h->cm_num);
+      cm_num = prop.data.colormap.num;
+      cm_map = (unsigned char *) SDL_realloc(h->cm_map, sizeof (unsigned char) * 3 * cm_num);
+      if (cm_map) {
+        h->cm_num = cm_num;
+        h->cm_map = cm_map;
+        SDL_memcpy (h->cm_map, prop.data.colormap.cmap, 3*sizeof (char)*h->cm_num);
+      }
       SDL_free (prop.data.colormap.cmap);
+
+      if (!cm_map) {
+        free_xcf_header(h);
+        return NULL;
+      }
     }
   } while (prop.id != PROP_END);
 
@@ -468,10 +483,14 @@ static unsigned char * load_xcf_tile_rle (SDL_RWops * src, Uint32 len, int bpp,
   int i, size, count, j, length;
   unsigned char val;
 
+  if (len == 0) {  /* probably bogus data. */
+    return NULL;
+  }
+
   t = load = (unsigned char *) SDL_malloc (len);
   reallen = SDL_RWread (src, t, 1, len);
 
-  data = (unsigned char *) SDL_malloc (x*y*bpp);
+  data = (unsigned char *) SDL_calloc (1, x*y*bpp);
   for (i = 0; i < bpp; i++) {
     d    = data + i;
     size = x*y;
@@ -482,38 +501,54 @@ static unsigned char * load_xcf_tile_rle (SDL_RWops * src, Uint32 len, int bpp,
 
       length = val;
       if (length >= 128) {
-    length = 255 - (length - 1);
-    if (length == 128) {
-      length = (*t << 8) + t[1];
-      t += 2;
-    }
+        length = 255 - (length - 1);
+        if (length == 128) {
+          length = (*t << 8) + t[1];
+          t += 2;
+        }
 
-    count += length;
-    size -= length;
+        if (((size_t) (t - load) + length) >= len) {
+          break;  /* bogus data */
+        } else if (length > size) {
+          break;  /* bogus data */
+        }
 
-    while (length-- > 0) {
-      *d = *t++;
-      d += bpp;
-    }
-      }
-      else {
-    length += 1;
-    if (length == 128) {
-      length = (*t << 8) + t[1];
-      t += 2;
-    }
+        count += length;
+        size -= length;
 
-    count += length;
-    size -= length;
+        while (length-- > 0) {
+          *d = *t++;
+          d += bpp;
+        }
+      } else {
+        length += 1;
+        if (length == 128) {
+          length = (*t << 8) + t[1];
+          t += 2;
+        }
 
-    val = *t++;
+        if (((size_t) (t - load)) >= len) {
+          break;  /* bogus data */
+        } else if (length > size) {
+          break;  /* bogus data */
+        }
 
-    for (j = 0; j < length; j++) {
-      *d = val;
-      d += bpp;
-    }
+        count += length;
+        size -= length;
+
+        val = *t++;
+
+        for (j = 0; j < length; j++) {
+          *d = val;
+          d += bpp;
+        }
       }
     }
+
+    if (size > 0) {
+      break;  /* just drop out, untouched data initialized to zero. */
+    }
+
   }
 
   SDL_free (load);
@@ -544,149 +579,162 @@ static void create_channel_surface (SDL_Surface * surf, xcf_image_type itype, Ui
   SDL_FillRect (surf, NULL, c);
 }
 
-static int do_layer_surface (SDL_Surface * surface, SDL_RWops * src, xcf_header * head, xcf_layer * layer, load_tile_type load_tile) {
-  xcf_hierarchy * hierarchy;
-  xcf_level     * level;
-  unsigned char * tile;
-  Uint8  * p8;
-  Uint16 * p16;
-  Uint32 * p;
-  int i, j;
-  Uint32 x, y, tx, ty, ox, oy;
-  Uint32 *row;
+static int 
+do_layer_surface(SDL_Surface * surface, SDL_RWops * src, xcf_header * head, xcf_layer * layer, load_tile_type load_tile)
+{
+    xcf_hierarchy  *hierarchy;
+    xcf_level      *level;
+    unsigned char  *tile;
+    Uint8          *p8;
+    Uint16         *p16;
+    Uint32         *p;
+    int            i, j;
+    Uint32         x, y, tx, ty, ox, oy;
+    Uint32         *row;
 
-  SDL_RWseek (src, layer->hierarchy_file_offset, RW_SEEK_SET);
-  hierarchy = read_xcf_hierarchy (src);
+    SDL_RWseek(src, layer->hierarchy_file_offset, RW_SEEK_SET);
+    hierarchy = read_xcf_hierarchy(src);
 
-  level = NULL;
-  for (i = 0; hierarchy->level_file_offsets [i]; i++) {
-    SDL_RWseek (src, hierarchy->level_file_offsets [i], RW_SEEK_SET);
-    level = read_xcf_level (src);
-
-    ty = tx = 0;
-    for (j = 0; level->tile_file_offsets [j]; j++) {
-      SDL_RWseek (src, level->tile_file_offsets [j], RW_SEEK_SET);
-      ox = tx+64 > level->width ? level->width % 64 : 64;
-      oy = ty+64 > level->height ? level->height % 64 : 64;
-
-      if (level->tile_file_offsets [j+1]) {
-    tile = load_tile
-      (src,
-       level->tile_file_offsets [j+1] - level->tile_file_offsets [j],
-       hierarchy->bpp,
-       ox, oy);
-      }
-      else {
-    tile = load_tile
-      (src,
-       ox*oy*6,
-       hierarchy->bpp,
-       ox, oy);
-      }
-
-      p8  = tile;
-      p16 = (Uint16 *) p8;
-      p   = (Uint32 *) p8;
-      for (y=ty; y < ty+oy; y++) {
-    row = (Uint32 *)((Uint8 *)surface->pixels + y*surface->pitch + tx*4);
-    switch (hierarchy->bpp) {
-    case 4:
-      for (x=tx; x < tx+ox; x++)
-        *row++ = Swap32 (*p++);
-      break;
-    case 3:
-      for (x=tx; x < tx+ox; x++) {
-        *row = 0xFF000000;
-        *row |= ((Uint32) *(p8++) << 16);
-        *row |= ((Uint32) *(p8++) << 8);
-        *row |= ((Uint32) *(p8++) << 0);
-        row++;
-      }
-      break;
-    case 2: // Indexed/Greyscale + Alpha
-      switch (head->image_type) {
-      case IMAGE_INDEXED:
-        for (x=tx; x < tx+ox; x++) {
-          *row =  ((Uint32) (head->cm_map [*p8*3])     << 16);
-          *row |= ((Uint32) (head->cm_map [*p8*3+1])   << 8);
-          *row |= ((Uint32) (head->cm_map [*p8++*3+2]) << 0);
-          *row |= ((Uint32) *p8++ << 24);;
-          row++;
-        }
-        break;
-      case IMAGE_GREYSCALE:
-        for (x=tx; x < tx+ox; x++) {
-          *row = ((Uint32) *p8 << 16);
-          *row |= ((Uint32) *p8 << 8);
-          *row |= ((Uint32) *p8++ << 0);
-          *row |= ((Uint32) *p8++ << 24);;
-          row++;
-        }
-        break;
-      default:
-        fprintf (stderr, "Unknown Gimp image type (%d)\n", head->image_type);
-        if (hierarchy)
-        {
-          if (hierarchy->level_file_offsets)
-            SDL_free(hierarchy->level_file_offsets);
-         
-          free_xcf_hierarchy(hierarchy);
-        }
-        if (level)     
-          free_xcf_level (level);
+    if (hierarchy->bpp > 4) {  /* unsupported. */
+        SDL_Log("Unknown Gimp image bpp (%u)\n", (unsigned int) hierarchy->bpp);
+        free_xcf_hierarchy(hierarchy);
         return 1;
-      }
-      break;
-    case 1: // Indexed/Greyscale
-      switch (head->image_type) {
-      case IMAGE_INDEXED:
-        for (x = tx; x < tx+ox; x++) {
-          *row++ = 0xFF000000
-        | ((Uint32) (head->cm_map [*p8*3]) << 16)
-        | ((Uint32) (head->cm_map [*p8*3+1]) << 8)
-        | ((Uint32) (head->cm_map [*p8*3+2]) << 0);
-          p8++;
-        }
-        break;
-      case IMAGE_GREYSCALE:
-        for (x=tx; x < tx+ox; x++) {
-          *row++ = 0xFF000000
-        | (((Uint32) (*p8)) << 16)
-        | (((Uint32) (*p8)) << 8)
-        | (((Uint32) (*p8)) << 0);
-            ++p8;
-        }
-        break;
-      default:
-        fprintf (stderr, "Unknown Gimp image type (%d)\n", head->image_type);
-        if (tile)
-          free_xcf_tile (tile);
-        if (level)
-          free_xcf_level (level);
-        if (hierarchy)
-          free_xcf_hierarchy (hierarchy);
+    }
+
+    if ((hierarchy->width > 20000) || (hierarchy->height > 20000)) {  /* arbitrary limit to avoid integer overflow. */
+        SDL_Log("Gimp image too large (%ux%u)\n", (unsigned int) hierarchy->width, (unsigned int) hierarchy->height);
+        free_xcf_hierarchy(hierarchy);
         return 1;
-      }
-      break;
     }
-      }
-      tx += 64;
-      if (tx >= level->width) {
-    tx = 0;
-    ty += 64;
-      }
-      if (ty >= level->height) {
-    break;
-      }
 
-      free_xcf_tile (tile);
+    level = NULL;
+    for (i = 0; hierarchy->level_file_offsets[i]; i++) {
+        SDL_RWseek(src, hierarchy->level_file_offsets[i], RW_SEEK_SET);
+        level = read_xcf_level(src);
+
+        ty = tx = 0;
+        for (j = 0; level->tile_file_offsets[j]; j++) {
+            SDL_RWseek(src, level->tile_file_offsets[j], RW_SEEK_SET);
+            ox = tx + 64 > level->width ? level->width % 64 : 64;
+            oy = ty + 64 > level->height ? level->height % 64 : 64;
+
+            if (level->tile_file_offsets[j + 1]) {
+                tile = load_tile(src, level->tile_file_offsets[j + 1] - level->tile_file_offsets[j], hierarchy->bpp, ox, oy);
+            } else {
+                tile = load_tile(src, ox * oy * 6, hierarchy->bpp, ox, oy);
+            }
+
+            if (!tile) {
+                if (hierarchy) {
+                    free_xcf_hierarchy(hierarchy);
+                }
+                if (level) {
+                    free_xcf_level(level);
+                }
+                return 1;
+            }
+
+            p8 = tile;
+            p16 = (Uint16 *) p8;
+            p = (Uint32 *) p8;
+            for (y = ty; y < ty + oy; y++) {
+                row = (Uint32 *) ((Uint8 *) surface->pixels + y * surface->pitch + tx * 4);
+                switch (hierarchy->bpp) {
+                case 4:
+                    for (x = tx; x < tx + ox; x++)
+                        *row++ = Swap32(*p++);
+                    break;
+                case 3:
+                    for (x = tx; x < tx + ox; x++) {
+                        *row = 0xFF000000;
+                        *row |= ((Uint32)*p8++ << 16);
+                        *row |= ((Uint32)*p8++ << 8);
+                        *row |= ((Uint32)*p8++ << 0);
+                        row++;
+                    }
+                    break;
+                case 2:
+                    /* Indexed / Greyscale + Alpha */
+                    switch (head->image_type) {
+                    case IMAGE_INDEXED:
+                        for (x = tx; x < tx + ox; x++) {
+                            *row = ((Uint32)(head->cm_map[*p8 * 3]) << 16);
+                            *row |= ((Uint32)(head->cm_map[*p8 * 3 + 1]) << 8);
+                            *row |= ((Uint32)(head->cm_map[*p8++ * 3 + 2]) << 0);
+                            *row |= ((Uint32)*p8++ << 24);
+                            row++;
+                        }
+                        break;
+                    case IMAGE_GREYSCALE:
+                        for (x = tx; x < tx + ox; x++) {
+                            *row = ((Uint32)*p8 << 16);
+                            *row |= ((Uint32)*p8 << 8);
+                            *row |= ((Uint32)*p8++ << 0);
+                            *row |= ((Uint32)*p8++ << 24);
+                            row++;
+                        }
+                        break;
+                    default:
+                        SDL_Log("Unknown Gimp image type (%d)\n", head->image_type);
+                        if (hierarchy) {
+                            free_xcf_hierarchy(hierarchy);
+                        }
+                        if (level)
+                            free_xcf_level(level);
+                        return 1;
+                    }
+                    break;
+                case 1:
+                    /* Indexed / Greyscale */
+                    switch (head->image_type) {
+                    case IMAGE_INDEXED:
+                        for (x = tx; x < tx + ox; x++) {
+                            *row++ = 0xFF000000
+                                | ((Uint32)(head->cm_map[*p8 * 3]) << 16)
+                                | ((Uint32)(head->cm_map[*p8 * 3 + 1]) << 8)
+                                | ((Uint32)(head->cm_map[*p8 * 3 + 2]) << 0);
+                            p8++;
+                        }
+                        break;
+                    case IMAGE_GREYSCALE:
+                        for (x = tx; x < tx + ox; x++) {
+                            *row++ = 0xFF000000
+                                | (((Uint32)(*p8)) << 16)
+                                | (((Uint32)(*p8)) << 8)
+                                | (((Uint32)(*p8)) << 0);
+                            ++p8;
+                        }
+                        break;
+                    default:
+                        SDL_Log("Unknown Gimp image type (%d)\n", head->image_type);
+                        if (tile)
+                            free_xcf_tile(tile);
+                        if (level)
+                            free_xcf_level(level);
+                        if (hierarchy)
+                            free_xcf_hierarchy(hierarchy);
+                        return 1;
+                    }
+                    break;
+                }
+            }
+            free_xcf_tile(tile);
+
+            tx += 64;
+            if (tx >= level->width) {
+                tx = 0;
+                ty += 64;
+            }
+            if (ty >= level->height) {
+                break;
+            }
+        }
+        free_xcf_level(level);
     }
-    free_xcf_level (level);
-  }
 
-  free_xcf_hierarchy (hierarchy);
+    free_xcf_hierarchy(hierarchy);
 
-  return 0;
+    return 0;
 }
 
 SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
@@ -702,7 +750,7 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
 
   unsigned char * (* load_tile) (SDL_RWops *, Uint32, int, int, int);
 
-  if ( !src ) {
+  if (!src) {
     /* The error message has been set in SDL_RWFromFile */
     return NULL;
   }
@@ -711,7 +759,10 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
   /* Initialize the data we will clean up when we're done */
   surface = NULL;
 
-  head = read_xcf_header (src);
+  head = read_xcf_header(src);
+  if (!head) {
+    return NULL;
+  }
 
   switch (head->compr) {
   case COMPR_NONE:
@@ -721,7 +772,7 @@ SDL_Surface *IMG_LoadXCF_RW(SDL_RWops *src)
     load_tile = load_xcf_tile_rle;
     break;
   default:
-    fprintf (stderr, "Unsupported Compression.\n");
+    SDL_Log("Unsupported Compression.\n");
     free_xcf_header (head);
     return NULL;
   }
